
/// TODO: let's add a common header in these examples, e.g., copyright + reference to the RAM article, how to cite, usaga information, etc.


/// CODE EXAMPLE IN THE ARTICLE ////////////////////////////////////////
// #include <pcl/surface/organized_fast_mesh.h>
// #include <pcl/registration/gicp.h>
// #include <pcl/features/from_meshes.h>
// PointCloud<PointT>::Ptr cloud;
// PolygonMesh::Ptr mesh;
// (...)
// // reconstruct meshes for source and target
// OrganizedFastMesh<PointT> fast_mesh;
// fast_mesh.setInputCloud(cloud);
// fast_mesh.reconstruct(*mesh);
// (...)
// // compute normals and covariances for source and target
// PointCloud<Normal>::Ptr normals;
// boost::shared_ptr< std::vector<Eigen::Vector3d> > covs;
// pcl::features::computeApproximateNormals(
// *cloud, mesh->polygons, *normals);
// pcl::features::computeApproximateCovariances(
// *cloud, normals, *covs);
// (...)
// // setup Generalized-ICP
// GeneralizedIterativeClosestPoint<PointT, PointT> gicp;
// gicp.setMaxCorrespondenceDistance(1);
// gicp.setInputSource(source_cloud);
// gicp.setInputTarget(target_cloud);
// gicp.setSourceCovariances(source_covariances);
// gicp.setTargetCovariances(target_covariances);
// // run registration and get transformation
// PointCloud<PointT> output;
// gicp.align(output);
// Eigen::Matrix4f transform = gicp.getFinalTransformation();
/// CODE EXAMPLE IN THE ARTICLE ////////////////////////////////////////

#include <pcl/console/parse.h>
#include <pcl/point_types.h>
#include <pcl/point_cloud.h>
#include <pcl/io/pcd_io.h>
#include <pcl/io/vtk_lib_io.h>

#include <pcl/surface/organized_fast_mesh.h>
#include <pcl/registration/gicp.h>
#include <pcl/features/from_meshes.h>
#include <pcl/surface/simplification_remove_unused_vertices.h>

typedef pcl::PointXYZ PointT;

bool readCloud(const std::string& filename, pcl::PointCloud<PointT>& cloud)
{
  PCL_INFO("Reading %s\n", filename.c_str());
  pcl::PCDReader reader;
  if (reader.read(filename, cloud) < 0)
  {
    PCL_ERROR("Error opening file: %s.\n", filename.c_str());
    return false;
  }
    
  if (cloud.points.empty())
  {
    PCL_ERROR("Error: The point cloud in %s is empty!\n", filename.c_str());
    return false;
  }

  if (!cloud.isOrganized())
  {
    PCL_ERROR("Error: The point cloud in %s is not organized!\n", filename.c_str());
    return false;
  }
  return true;
}

int main(int argc, char** argv)
{
  // get filenames from command line
  std::vector<int> input_file_indices;
  std::vector<int> pcd_file_indices = pcl::console::parse_file_extension_argument (argc, argv, ".pcd");
  if (pcd_file_indices.size() < 2)
  {
    PCL_ERROR("USAGE: %s source.pcd target.pcd\n", argv[0]);
    exit(1);
  }

  // read source and target point cloud from .pcd filenames
  pcl::PointCloud<PointT>::Ptr cloud_source (new pcl::PointCloud<PointT>);
  pcl::PointCloud<PointT>::Ptr cloud_target (new pcl::PointCloud<PointT>);
  if (!readCloud(argv[pcd_file_indices[0]], *cloud_source))
    exit(1);
  if (!readCloud(argv[pcd_file_indices[1]], *cloud_target))
    exit(1);

  // compute approximate surface reconstruction
  pcl::PolygonMesh::Ptr mesh_source(new pcl::PolygonMesh);
  pcl::PolygonMesh::Ptr mesh_target(new pcl::PolygonMesh);
  pcl::OrganizedFastMesh<PointT> fast_mesh;
  fast_mesh.setTriangulationType(fast_mesh.QUAD_MESH);
  fast_mesh.setTrianglePixelSizeRows(1);       
  fast_mesh.setTrianglePixelSizeColumns(4); 
  fast_mesh.setMaxEdgeLength(0.0f, 2 * atan(10.0f * M_PI / 180.0f), 0.0f);
  fast_mesh.setInputCloud(cloud_source);
  fast_mesh.reconstruct(*mesh_source);
  fast_mesh.setInputCloud(cloud_target);
  fast_mesh.reconstruct(*mesh_target);

  // remove unused vertices
  pcl::surface::SimplificationRemoveUnusedVertices filter_vertices;
  pcl::PolygonMesh::Ptr mesh_filtered(new pcl::PolygonMesh());
  std::vector<int> original_indices;
  filter_vertices.simplify(*mesh_source, *mesh_filtered, original_indices);
  mesh_source.swap(mesh_filtered);
  filter_vertices.simplify(*mesh_target, *mesh_filtered, original_indices);
  mesh_target.swap(mesh_filtered);

  // retrieve points from mesh (with unused vertices removed)
  pcl::fromPCLPointCloud2(mesh_source->cloud, *cloud_source);
  pcl::fromPCLPointCloud2(mesh_target->cloud, *cloud_target);

  // compute normals and covariances for source and target
  pcl::PointCloud<pcl::Normal>::Ptr normals_source(new pcl::PointCloud<pcl::Normal>);
  pcl::PointCloud<pcl::Normal>::Ptr normals_target(new pcl::PointCloud<pcl::Normal>);
  pcl::GeneralizedIterativeClosestPoint<PointT, PointT>::MatricesVectorPtr covariances_source(new pcl::GeneralizedIterativeClosestPoint<PointT, PointT>::MatricesVector);
  pcl::GeneralizedIterativeClosestPoint<PointT, PointT>::MatricesVectorPtr covariances_target(new pcl::GeneralizedIterativeClosestPoint<PointT, PointT>::MatricesVector);
  pcl::features::computeApproximateNormals(*cloud_source, mesh_source->polygons, *normals_source);
  pcl::features::computeApproximateNormals(*cloud_target, mesh_target->polygons, *normals_target);
  pcl::features::computeApproximateCovariances(*cloud_source, *normals_source, *covariances_source);
  pcl::features::computeApproximateCovariances(*cloud_target, *normals_target, *covariances_target);

  // setup Generalized-ICP
  pcl::GeneralizedIterativeClosestPoint<PointT, PointT> gicp;
  gicp.setMaxCorrespondenceDistance(2.5);
  gicp.setCorrespondenceRandomness(5);
  gicp.setInputSource(cloud_source);
  gicp.setInputTarget(cloud_target);
  gicp.setSourceCovariances(covariances_source);
  gicp.setTargetCovariances(covariances_target);

  // run registration and get transformation
  pcl::PointCloud<PointT> output;
  gicp.align(output);
  Eigen::Matrix4f transform = gicp.getFinalTransformation();

  // output result
  pcl::io::savePCDFileBinary("output.pcd", output);
  std::cout << "Computed transformation matrix:" << std::endl << transform << std::endl;

  return 0;
}


